@@ -15,15 +15,12 @@
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
 import logging
-import sys
+from typing import Any, List, Optional
 
-from gi.repository import Gio
-from gi.repository import GLib
-from gi.repository import GObject
-from gi.repository import GtkSource
+from gi.repository import Gio, GLib, GObject, GtkSource
 
 from meld.conf import _
-from meld.settings import bind_settings, meldsettings
+from meld.settings import bind_settings
 
 log = logging.getLogger(__name__)
 
@@ -41,12 +38,6 @@ class MeldBuffer(GtkSource.Buffer):
         bind_settings(self)
         self.data = MeldBufferData()
         self.undo_sequence = None
-        meldsettings.connect('changed', self.on_setting_changed)
-        self.set_style_scheme(meldsettings.style_scheme)
-
-    def on_setting_changed(self, meldsettings, key):
-        if key == 'style-scheme':
-            self.set_style_scheme(meldsettings.style_scheme)
 
     def do_begin_user_action(self, *args):
         if self.undo_sequence:
@@ -87,14 +78,14 @@ class MeldBuffer(GtkSource.Buffer):
 
 class MeldBufferData(GObject.GObject):
 
-    __gsignals__ = {
-        str('file-changed'): (GObject.SignalFlags.RUN_FIRST, None, ()),
-    }
+    @GObject.Signal('file-changed')
+    def file_changed_signal(self) -> None:
+        ...
 
     encoding = GObject.Property(
         type=GtkSource.Encoding,
         nick="The file encoding of the linked GtkSourceFile",
-        default=None,
+        default=GtkSource.Encoding.get_utf8(),
     )
 
     def __init__(self):
@@ -160,7 +151,7 @@ class MeldBufferData(GObject.GObject):
     def _handle_file_change(self, monitor, f, other_file, event_type):
         mtime = self._query_mtime(f)
         if self._disk_mtime and mtime and mtime > self._disk_mtime:
-            self.emit('file-changed')
+            self.file_changed_signal.emit()
         self._disk_mtime = mtime or self._disk_mtime
 
     @property
@@ -224,76 +215,80 @@ class BufferLines:
     This class allows a Gtk.TextBuffer to be treated as a list of lines of
     possibly-filtered text. If no filter is given, the raw output from the
     Gtk.TextBuffer is used.
-
-    The logic here (and in places in FileDiff) requires that Python's
-    unicode splitlines() implementation and Gtk.TextBuffer agree on where
-    linebreaks occur. Happily, this is usually the case.
     """
 
-    def __init__(self, buf, textfilter=None):
+    #: Cached copy of the (possibly filtered) text in a single line,
+    #: where an entry of None indicates that there is no cached result
+    #: available.
+    lines: List[Optional[str]]
+
+    def __init__(self, buf, textfilter=None, *, cache_debug: bool = False):
         self.buf = buf
         if textfilter is not None:
             self.textfilter = textfilter
         else:
             self.textfilter = lambda x, buf, start_iter, end_iter: x
 
+        self.lines = [None] * self.buf.get_line_count()
+        self.mark = buf.create_mark(
+            "bufferlines-insert", buf.get_start_iter(), True,
+        )
+
+        buf.connect("insert-text", self.on_insert_text),
+        buf.connect("delete-range", self.on_delete_range),
+        buf.connect_after("insert-text", self.after_insert_text),
+        if cache_debug:
+            buf.connect_after("insert-text", self._check_cache_invariant)
+            buf.connect_after("delete-range", self._check_cache_invariant)
+
+    def _check_cache_invariant(self, *args: Any) -> None:
+        if len(self.lines) != len(self):
+            log.error(
+                "Cache line count does not match buffer line count: "
+                f"{len(self.lines)} != {len(self)}",
+            )
+
+    def on_insert_text(self, buf, it, text, textlen):
+        buf.move_mark(self.mark, it)
+
+    def after_insert_text(self, buf, it, newtext, textlen):
+        start_idx = buf.get_iter_at_mark(self.mark).get_line()
+        end_idx = it.get_line() + 1
+        # Replace the insertion-point cache line with a list of empty
+        # lines. In the single-line case this will be a single element
+        # substitution; for multi-line inserts, we will replace the
+        # single insertion point line with several empty cache lines.
+        self.lines[start_idx:start_idx + 1] = [None] * (end_idx - start_idx)
+
+    def on_delete_range(self, buf, it0, it1):
+        start_idx = it0.get_line()
+        end_idx = it1.get_line() + 1
+        self.lines[start_idx:end_idx] = [None]
+
     def __getitem__(self, key):
         if isinstance(key, slice):
             lo, hi, _ = key.indices(self.buf.get_line_count())
 
-            # FIXME: If we ask for arbitrary slices past the end of the buffer,
-            # this will return the last line.
-            start = self.buf.get_iter_at_line_or_eof(lo)
-            end = self.buf.get_iter_at_line_or_eof(hi)
-            txt = self.buf.get_text(start, end, False)
+            for idx in range(lo, hi):
+                if self.lines[idx] is None:
+                    self.lines[idx] = self[idx]
 
-            filter_txt = self.textfilter(txt, self.buf, start, end)
-            lines = filter_txt.splitlines()
-            ends = filter_txt.splitlines(True)
-
-            # The last line in a Gtk.TextBuffer is guaranteed never to end in a
-            # newline. As splitlines() discards an empty line at the end, we
-            # need to artificially add a line if the requested slice is past
-            # the end of the buffer, and the last line in the slice ended in a
-            # newline.
-            if hi >= self.buf.get_line_count() and \
-               lo < self.buf.get_line_count() and \
-               (len(lines) == 0 or len(lines[-1]) != len(ends[-1])):
-                lines.append("")
-                ends.append("")
-
-            hi = self.buf.get_line_count() if hi == sys.maxsize else hi
-            if hi - lo != len(lines):
-                # These codepoints are considered line breaks by Python, but
-                # not by GtkTextStore.
-                additional_breaks = set(('\x0b', '\x0c', '\x85', '\u2028'))
-                i = 0
-                while i < len(ends):
-                    line, end = lines[i], ends[i]
-                    # It's possible that the last line in a file would end in a
-                    # line break character, which requires no joining.
-                    if end and end[-1] in additional_breaks and \
-                       (not line or line[-1] not in additional_breaks):
-                        assert len(ends) >= i + 1
-                        lines[i:i + 2] = [line + end[-1] + lines[i + 1]]
-                        ends[i:i + 2] = [end + ends[i + 1]]
-                    else:
-                        # We only increment if we don't correct a line, to
-                        # handle the case of a single line having multiple
-                        # additional_breaks characters that need correcting.
-                        i += 1
-
-            return lines
+            return self.lines[lo:hi]
 
         elif isinstance(key, int):
             if key >= len(self):
                 raise IndexError
-            line_start = self.buf.get_iter_at_line_or_eof(key)
-            line_end = line_start.copy()
-            if not line_end.ends_line():
-                line_end.forward_to_line_end()
-            txt = self.buf.get_text(line_start, line_end, False)
-            return self.textfilter(txt, self.buf, line_start, line_end)
+
+            if self.lines[key] is None:
+                line_start = self.buf.get_iter_at_line_or_eof(key)
+                line_end = line_start.copy()
+                if not line_end.ends_line():
+                    line_end.forward_to_line_end()
+                txt = self.buf.get_text(line_start, line_end, False)
+                txt = self.textfilter(txt, self.buf, line_start, line_end)
+                self.lines[key] = txt
+
+            return self.lines[key]
 
     def __len__(self):
         return self.buf.get_line_count()
