@@ -29,9 +29,9 @@ import re
 import shutil
 import subprocess
 import tempfile
+from typing import ClassVar
 
-from gi.repository import Gio
-from gi.repository import GLib
+from gi.repository import Gio, GLib
 
 from meld.conf import _
 
@@ -42,7 +42,7 @@ from meld.conf import _
 (STATE_IGNORED, STATE_NONE, STATE_NORMAL, STATE_NOCHANGE,
     STATE_ERROR, STATE_EMPTY, STATE_NEW,
     STATE_MODIFIED, STATE_RENAMED, STATE_CONFLICT, STATE_REMOVED,
-    STATE_MISSING, STATE_NONEXIST, STATE_MAX) = list(range(14))
+    STATE_MISSING, STATE_NONEXIST, STATE_SPINNER, STATE_MAX) = list(range(15))
 
 # VC conflict types
 (CONFLICT_MERGED, CONFLICT_BASE, CONFLICT_LOCAL,
@@ -79,6 +79,7 @@ class Entry:
         STATE_REMOVED: _("Removed"),
         STATE_MISSING: _("Missing"),
         STATE_NONEXIST: _("Not present"),
+        STATE_SPINNER: _("Scanning…"),
     }
 
     def __init__(self, path, name, state, isdir, options=None):
@@ -126,12 +127,12 @@ class Entry:
 
 class Vc:
 
-    VC_DIR = None
+    VC_DIR: ClassVar[str]
 
     #: Whether to walk the current location's parents to find a
     #: repository root. Only used in legacy version control systems
     #: (e.g., old SVN, CVS, RCS).
-    VC_ROOT_WALK = True
+    VC_ROOT_WALK: ClassVar[bool] = True
 
     def __init__(self, path):
         # Save the requested comparison location. The location may be a
@@ -182,7 +183,7 @@ class Vc:
         version control system has a (possibly partial) pre-filled
         message, or None if no such message exists.
 
-        This method should use pre-filled commit messages whereever
+        This method should use pre-filled commit messages wherever
         provided by the version control system, most commonly these are
         given in merging, revert or cherry-picking scenarios.
         """
@@ -231,8 +232,8 @@ class Vc:
             valid_actions.add('add')
         if all(s == STATE_CONFLICT for s in states):
             valid_actions.add('resolve')
-        if (all(s not in non_removeable_states for s in states)
-                and self.root not in path_states.keys()):
+        if (all(s not in non_removeable_states for s in states) and
+                self.root not in path_states.keys()):
             valid_actions.add('remove')
         if all(s not in non_revertable_states for s in states):
             valid_actions.add('revert')
@@ -325,7 +326,7 @@ class Vc:
         except GLib.Error as e:
             if e.domain != 'g-io-error-quark':
                 raise
-            # Handling for non-existant files (or other IO errors)
+            # Handling for non-existent files (or other IO errors)
             name = path
             isdir = False
 
@@ -413,7 +414,7 @@ def popen(cmd, cwd=None, use_locale_encoding=True):
     return process.stdout
 
 
-def call_temp_output(cmd, cwd, file_id=''):
+def call_temp_output(cmd, cwd, file_id='', suffix=None):
     """Call `cmd` in `cwd` and write the output to a temporary file
 
     This returns the name of the temporary file used. It is the
@@ -421,6 +422,9 @@ def call_temp_output(cmd, cwd, file_id=''):
 
     If `file_id` is provided, it is used as part of the
     temporary file's name, for ease of identification.
+
+    If `suffix` is provided, it is used as the extension
+    of the temporary file's name.
     """
     process = subprocess.Popen(
         cmd, cwd=cwd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
@@ -430,7 +434,8 @@ def call_temp_output(cmd, cwd, file_id=''):
     # sane response is to return an empty temp file.
 
     prefix = 'meld-tmp' + ('-' + file_id if file_id else '')
-    with tempfile.NamedTemporaryFile(prefix=prefix, delete=False) as f:
+    with tempfile.NamedTemporaryFile(prefix=prefix,
+                                     suffix=suffix, delete=False) as f:
         shutil.copyfileobj(vc_file, f)
     return f.name
 
